import QtQuick 2.0
import Ubuntu.Components 0.1
import Ubuntu.Components.ListItems 0.1 as ListItem
import Ubuntu.Components.Popups 0.1

import "board.js" as Board
import "Constants.js" as Constants

Grid {
    property int cellSide: Math.min(parent.width, parent.height) / 8
    property bool lock: false
    anchors.centerIn: parent
    rows: 8
    property int cols: 8
    clip: false

    //Animate the movement of the piece in the grid
    function movePiece(fromRow, fromCol, toRow, toCol){
        repeater.itemAt(fromRow * rows + fromCol).moveCircle(toRow, toCol);
    }

    //Handle the click event
    function handleClick(x, y){
        //Discard user interaction during animations
        if(checkboardui.lock) return

        var computer = Board.PlayerTwo;
        var game = Board.game;

        x = x - checkboardui.x;
        y = y - checkboardui.y;

        //Check if inside the board
        if(x < 0 || x > checkboardui.width)  return
        if(y < 0 || y > checkboardui.height) return

        //Check if the pc is playing
        if(game.player() === computer) return

        var row = Math.floor(y / checkboardui.cellSide);
        var col = Math.floor(x / checkboardui.cellSide);

        var player = game.playerAt(row, col);
        if(player){
            if(player !== game.player()) return

            //If the player has selected his own piece update view
            if(!game.inMultiTurn()){
                Board.selectedRow = row;
                Board.selectedCol = col;
                checkboardui.draw(game.board);
                return;
            }
        }

        if(Board.selectedRow === undefined) return;

        var fromRow = Board.selectedRow;
        var fromCol = Board.selectedCol;
        if(!game.inMultiTurn()){
            Board.selectedCol = undefined;
            Board.selectedRow = undefined;
        }

        try{
            var winner = game.move(fromRow, fromCol, row, col);
            checkboardui.movePiece(fromRow, fromCol, row, col);
            if(winner){
                checkboardui.gamewon(winner);
                return;
            }
        }catch(e){
            console.log(e);
            checkboardui.draw(game.board);
            return;
        }

        if(Board.game.inMultiTurn()){
            Board.selectedRow = row;
            Board.selectedCol = col;
            return;
        }

        //Start the movement of the AI
        if(Board.Checkers){
            aitimer.interval = 1000;
            aitimer.start();
        }
    }

    function start(){
        draw(Board.game.board);
    }

    function drawCircle(row, col, color, king){
        repeater.itemAt(row * rows + col).showCircle(color, king);
    }

    function hideCircle(row, col){
        repeater.itemAt(row * rows + col).hideCircle();
    }

    //Fire the animation of the selected cell
    function drawSelectedCell(row, col){
        repeater.itemAt(row * rows + col).onSelected();
    }

    //Stop the animation
    function stopSelectedCell(row, col){
        repeater.itemAt(row * rows + col).onNotSelected();
    }

    //The game is finished: notify the winner
    function gamewon(winner){
        var color;
        if(winner === 2) color = Constants.blackCheckerColor;
        else color = Constants.whiteCheckerColor;

        PopupUtils.open(Qt.resolvedUrl("EndOfGameDialog.qml"), checkboardui, {
                            title: i18n.tr("The Winner is..."),
                            checkercolor: color
                            });
    }

    //Draw the board
    function draw(b){
        for(var row = 0; row < rows; row++){
            for(var col = 0; col < cols; col++){
                var piece = b.getPlayer(row, col);

                if(piece === Board.Checkers.EmptyPlayer){
                    hideCircle(row, col);
                    continue;
                }

                if(row === Board.selectedRow && col === Board.selectedCol){
                    drawSelectedCell(row, col);
                } else {
                    stopSelectedCell(row, col);
                }

                var color = (piece === Board.Checkers.PlayerOne)
                        ? Constants.blackCheckerColor
                        : Constants.whiteCheckerColor;
                drawCircle(row, col, color, b.isKing(row, col));
            }
        }
    }

    Repeater {
        id: repeater
        model: 8*8

        Item{
            id: content
            width: checkboardui.cellSide; height: checkboardui.cellSide;

            //Animation used for the movement
            ParallelAnimation{
                id: movepieceanim
                NumberAnimation {id: xtras; property: "x"; duration: 400 }
                NumberAnimation {id: ytras; property: "y"; duration: 400 }

                onRunningChanged: {
                    if(movepieceanim.running === true){
                        content.z = 1;
                        checkboardui.lock = true;
                    } else {
                        checkboardui.lock = false;
                        var target = xtras.target;
                        target.opacity = 0;
                        target.x = xtras.from;
                        target.y = ytras.from;
                        content.z = 0;
                        checkboardui.draw(Board.game.board);
                    }
                }
            }

            //Animation used to differenciate the selected cell
            SequentialAnimation{
                id: selectanimation
                NumberAnimation {id: zoom;   property: "scale"; from: 1; to: 1.4; duration: 500; easing.type: Easing.InOutQuad }
                NumberAnimation {id: dezoom; property: "scale"; from: 1.4; to: 1; duration: 500; easing.type: Easing.InOutQuad }
                loops: Animation.Infinite
            }

            //Rectangle representing the cell of the board
            Rectangle {
                width: checkboardui.cellSide; height: checkboardui.cellSide
                property string defaultColor: getDefaultColor();
                color: defaultColor;
                //Return the color of the cell given the index
                function getDefaultColor(){
                    var row = index % 8
                    var col = Math.floor(index / 8)
                    if((row % 2 + col % 2) % 2 == 0) return Constants.brightCellColor
                    else return Constants.darkCellColor
                }
            }

            //Animate the given circle to another position
            function moveCircle(toRow, toCol){
                onNotSelected();
                xtras.target = circle;
                ytras.target = circle;
                xtras.from = circle.x;
                ytras.from = circle.y;
                xtras.to = (toCol - index % 8) * width;
                ytras.to = (toRow - Math.floor(index / 8)) * width;
                movepieceanim.start();
            }

            function showCircle(color, king){
                circle.opacity = 100;
                circle.border.color = color;
                innercircle.color = color;
                innercircle.opacity = 100;
                if(king)kingimg.opacity = 100;
                else kingimg.opacity = 0;
            }

            function hideCircle(){
                circle.opacity = 0;
                innercircle.opacity = 0;
            }

            function onNotSelected(){
                innercircle.scale = 1;
                selectanimation.stop();
            }

            function onSelected(){
                zoom.target = innercircle;
                dezoom.target = innercircle;
                selectanimation.start();
            }


            //Those are the actual circles
            Rectangle {
                id: circle
                antialiasing: true

                 width: parent.width * 0.9
                 x: width * 0.05
                 y: width * 0.05
                 clip: false
                 height: width
                 radius: width * 0.5
                 opacity: 0
                 color: "#000000ff"
                 border.width: width * 0.1

                 Rectangle{
                     id: innercircle
                     antialiasing: true
                    width:   parent.width * 0.6
                    anchors.centerIn: parent
                    height: width
                    radius: width * 0.6
                    opacity: 0

                    Image {
                        id: kingimg
                        antialiasing: true
                        source: "king.png"
                        anchors.centerIn: parent
                        width: parent.width * 0.8
                        height: parent.width * 0.8
                        opacity: 0
                    }
                }
            }
        }
    }

    Component.onCompleted: {
        var game = new Board.Checkers.Game(checkboardui);
        game.start();
    }
}
